/*-----------------------------------------------------------------------*/
/* Recurring events admin style */
/*-----------------------------------------------------------------------*/

.joy-rr__advanced-repeat,
.joy-rr__advanced-monthly-repeat,
.joy-rr__advanced-yearly-repeat {
    display:                 none;
}
.joy-rr__show-advanced label {
    display:                 block;
    text-align:              right;
    width:                   100%;
    float:                   none;
    font-weight:             normal;
    color:                   #43b1b0;
}
.joy-rr__short-input input {
    width:                   10em;
}
.joy-rr__days-input label {
    width:                   14%;
}
.joy-rr__double-input {
}
.joy-rr__half-field {
    display:                 inline-block;
    width:                   40%;
}
.joy-rr__half-field .field-content {
    width:                   100%;
}
.joy-rr__half-field:last-child {
    margin-left:             2.5%;
}
.joy-rr__ical-value {
    font-size:               10px;
    color:                   #aaa;
}
#id_overrides-chooser ul.actions {
    display:                 none;
}
