// Generated by CoffeeScript 1.10.0
(function() {
  var $, allHolidaysChanged, ref, ref1;

  $ = (ref = (ref1 = this.joyJQ) != null ? ref1 : this.$) != null ? ref : django.jQuery;

  allHolidaysChanged = function() {
    var allHols, fromBox, fromFilter, toBox;
    allHols = $("#id_all_holidays").prop("checked");
    if (allHols) {
      $("li.multiple_choice_field").hide();
    } else {
      $("li.multiple_choice_field").show();
      fromBox = $("#id_closed_for_from");
      fromFilter = $("#id_closed_for_filter");
      toBox = $("#id_closed_for_to");
      toBox.height(fromFilter.outerHeight() + fromBox.outerHeight());
    }
  };

  $(function() {
    allHolidaysChanged();
    $("#id_all_holidays").click(function(ev) {
      allHolidaysChanged();
      return true;
    });
  });

}).call(this);
